#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library
import logging
logger = logging.getLogger('PyAMS (utils)')

import threading
import venusian

# import interfaces
from pyramid.interfaces import INewRequest
from zope.component.interfaces import ComponentLookupError, ISite
from zope.traversing.interfaces import IBeforeTraverseEvent

# import packages
from pyramid.events import subscriber
from pyramid.threadlocal import manager
from zope.interface import implementedBy, providedBy


class LocalRegistry(threading.local):
    """Local registry"""

    _registry = None

    def get_registry(self):
        return self._registry

    def set_registry(self, registry):
        self._registry = registry

local_registry = LocalRegistry()


def get_local_registry():
    """Get local registry"""
    return local_registry.get_registry()


def set_local_registry(registry):
    """Define local registry"""
    local_registry.set_registry(registry)


@subscriber(INewRequest)
def handle_new_request(event):
    set_local_registry(None)


@subscriber(IBeforeTraverseEvent, context_selector=ISite)
def handle_site_before_traverse(event):
    """Define local registry when an object implementing ISite is traversed"""
    set_local_registry(event.object.getSiteManager())


def get_registries():
    """Get iterator of components registries"""
    registry = local_registry.get_registry()
    if registry is not None:
        yield registry
    for entry in reversed(manager.stack):
        registry = entry.get('registry')
        if registry is not None:
            yield registry


def registered_utilities():
    """Get utilities registrations as generator"""
    for registry in get_registries():
        for utility in registry.registeredUtilities():
            yield utility


def query_utility(provided, name='', default=None):
    """Query utility registered with given interface"""
    for registry in get_registries():
        utility = registry.queryUtility(provided, name, default)
        if utility is not None:
            return utility
    return default


def get_utility(provided, name=''):
    """Get utility registered with given interface"""
    for registry in get_registries():
        utility = registry.queryUtility(provided, name)
        if utility is not None:
            return utility
    raise ComponentLookupError(provided, name)


def get_utilities_for(interface):
    """Get utilities registered with given interface as (name, util)"""
    for registry in get_registries():
        for utility in registry.getUtilitiesFor(interface):
            yield utility


def get_all_utilities_registered_for(interface):
    """Get list of registered utilities for given interface"""
    result = []
    for registry in get_registries():
        for utilities in registry.getAllUtilitiesRegisteredFor(interface):
            result.append(utilities)
    return result


class utility_config(object):
    """Function or class decorator to declare a utility"""

    venusian = venusian

    def __init__(self, **settings):
        self.__dict__.update(settings)

    def __call__(self, wrapped):
        settings = self.__dict__.copy()
        depth = settings.pop('_depth', 0)

        def callback(context, name, ob):
            if type(ob) is type:
                factory = ob
                component = None
            else:
                factory = None
                component = ob

            provides = settings.get('provides')
            if provides is None:
                if factory:
                    provides = list(implementedBy(factory))
                else:
                    provides = list(providedBy(component))
                if len(provides) == 1:
                    provides = provides[0]
                else:
                    raise TypeError("Missing 'provides' argument")

            config = context.config.with_package(info.module)
            logger.debug("Registering utility {0} named '{1}' providing {2}".format(
                str(component) if component else str(factory),
                settings.get('name', ''),
                str(provides)))
            config.registry.registerUtility(component=component, factory=factory,
                                            provided=provides, name=settings.get('name', ''))

        info = self.venusian.attach(wrapped, callback, category='pyams_utility',
                                    depth=depth + 1)

        if info.scope == 'class':
            # if the decorator was attached to a method in a class, or
            # otherwise executed at class scope, we need to set an
            # 'attr' into the settings if one isn't already in there
            if settings.get('attr') is None:
                settings['attr'] = wrapped.__name__

        settings['_info'] = info.codeinfo  # fbo "action_method"
        return wrapped
